/*
 * Project: MoleCuilder
 * Description: creates and alters molecular systems
 * Copyright (C)  2013 Frederik Heber. All rights reserved.
 * 
 *
 *   This file is part of MoleCuilder.
 *
 *    MoleCuilder is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    MoleCuilder is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with MoleCuilder.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * FitPartialChargesAction.cpp
 *
 *  Created on: Jul 03, 2013
 *      Author: heber
 */

// include config.h
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// needs to come before MemDebug due to placement new
#include <boost/archive/text_iarchive.hpp>

#include "CodePatterns/MemDebug.hpp"

#include "Atom/atom.hpp"
#include "CodePatterns/Log.hpp"
#include "Fragmentation/Exporters/ExportGraph_ToFiles.hpp"
#include "Fragmentation/Graph.hpp"
#include "World.hpp"

#include <boost/bind.hpp>
#include <boost/filesystem.hpp>
#include <boost/foreach.hpp>
#include <algorithm>
#include <functional>
#include <iostream>
#include <string>

#include "Actions/PotentialAction/FitPartialChargesAction.hpp"

#include "Potentials/PartialNucleiChargeFitter.hpp"

#include "Element/element.hpp"
#include "Element/periodentafel.hpp"
#include "Fragmentation/Homology/AtomFragmentsMap.hpp"
#include "Fragmentation/Homology/HomologyContainer.hpp"
#include "Fragmentation/Homology/HomologyGraph.hpp"
#include "Fragmentation/Summation/SetValues/SamplingGrid.hpp"
#include "FunctionApproximation/Extractors.hpp"
#include "Potentials/PartialNucleiChargeFitter.hpp"
#include "Potentials/Particles/ParticleFactory.hpp"
#include "Potentials/Particles/ParticleRegistry.hpp"
#include "Potentials/SerializablePotential.hpp"
#include "World.hpp"

using namespace MoleCuilder;

// and construct the stuff
#include "FitPartialChargesAction.def"
#include "Action_impl_pre.hpp"
/** =========== define the function ====================== */

inline
HomologyGraph getFirstGraphwithSpecifiedElements(
    const HomologyContainer &homologies,
    const SerializablePotential::ParticleTypes_t &types)
{
  ASSERT( !types.empty(),
      "getFirstGraphwithSpecifiedElements() - charges is empty?");
  // create charges
  Fragment::charges_t charges;
  charges.resize(types.size());
  std::transform(types.begin(), types.end(),
      charges.begin(), boost::lambda::_1);
  // convert into count map
  Extractors::elementcounts_t counts_per_charge =
      Extractors::_detail::getElementCounts(charges);
  ASSERT( !counts_per_charge.empty(),
      "getFirstGraphwithSpecifiedElements() - charge counts are empty?");
  LOG(2, "DEBUG: counts_per_charge is " << counts_per_charge << ".");
  // we want to check each (unique) key only once
  HomologyContainer::const_key_iterator olditer = homologies.key_end();
  for (HomologyContainer::const_key_iterator iter =
      homologies.key_begin(); iter != homologies.key_end();
      iter = homologies.getNextKey(iter)) {
    // if it's the same as the old one, skip it
    if (olditer == iter)
      continue;
    else
      olditer = iter;
    // if it's a new key, check if every element has the right number of counts
    Extractors::elementcounts_t::const_iterator countiter = counts_per_charge.begin();
    for (; countiter != counts_per_charge.end(); ++countiter)
      if (!(*iter).hasTimesAtomicNumber(
          static_cast<size_t>(countiter->first),
          static_cast<size_t>(countiter->second))
          )
        break;
    if( countiter == counts_per_charge.end())
      return *iter;
  }
  return HomologyGraph();
}

void enforceZeroTotalCharge(
    PartialNucleiChargeFitter::charges_t &_averaged_charges)
{
  double charge_sum = 0.;
  charge_sum = std::accumulate(_averaged_charges.begin(), _averaged_charges.end(), charge_sum);
  if (fabs(charge_sum) > MYEPSILON) {
    std::transform(
        _averaged_charges.begin(), _averaged_charges.end(),
        _averaged_charges.begin(),
        boost::bind(std::minus<double>(), _1, charge_sum/_averaged_charges.size()));
  }
  charge_sum = 0.;
  charge_sum = std::accumulate(_averaged_charges.begin(), _averaged_charges.end(), charge_sum);
  ASSERT( fabs(charge_sum) < MYEPSILON,
      "enforceZeroTotalCharge() - enforcing neutral net charge failed, "
      +toString(charge_sum)+" is the remaining net charge.");
}

size_t obtainAverageChargesOverFragments(
    PartialNucleiChargeFitter::charges_t &_average_charges,
    const std::pair<
      HomologyContainer::const_iterator,
      HomologyContainer::const_iterator> &_range,
    const double _radius
    )
{
  HomologyContainer::const_iterator iter = _range.first;
  if (!iter->second.containsGrids) {
    ELOG(1, "This HomologyGraph does not contain sampled grids.");
    return 0;
  }
  _average_charges.resize(iter->second.fragment.getCharges().size(), 0.);
  size_t NoFragments = 0;
  for (;
      iter != _range.second; ++iter, ++NoFragments) {
    if (!iter->second.containsGrids) {
      ELOG(2, "This HomologyGraph does not contain sampled grids,\ndid you forget to add '--store-grids 1' to AnalyseFragmentResults.");
      return 0;
    }
    const Fragment &fragment = iter->second.fragment;
  //  const double &energy = iter->second.energy;
  //  const SamplingGrid &charge = iter->second.charge_distribution;
    const SamplingGrid &potential = iter->second.potential_distribution;
    if ((potential.level == 0)
        || ((potential.begin[0] == potential.end[0])
            && (potential.begin[1] == potential.end[1])
            && (potential.begin[2] == potential.end[2]))) {
      ELOG(1, "Sampled grid contains grid made of zero points.");
      return 0;
    }

    // then we extract positions from fragment
    PartialNucleiChargeFitter::positions_t positions;
    Fragment::positions_t fragmentpositions = fragment.getPositions();
    positions.reserve(fragmentpositions.size());
    BOOST_FOREACH( Fragment::position_t pos, fragmentpositions) {
      positions.push_back( Vector(pos[0], pos[1], pos[2]) );
    }
    PartialNucleiChargeFitter fitter(potential, positions, _radius);
    fitter();
    PartialNucleiChargeFitter::charges_t return_charges = fitter.getSolutionAsCharges_t();
    LOG(2, "DEBUG: fitted charges are " << return_charges);
    std::transform(
        return_charges.begin(), return_charges.end(),
        _average_charges.begin(),
        _average_charges.begin(),
        std::plus<PartialNucleiChargeFitter::charge_t>());
  }
  if (NoFragments != 0)
    std::transform(_average_charges.begin(), _average_charges.end(),
        _average_charges.begin(),
        std::bind1st(std::multiplies<PartialNucleiChargeFitter::charge_t>(),1./(double)NoFragments)
    );
  LOG(2, "DEBUG: final averaged charges are " << _average_charges);

  return NoFragments;
}

ActionState::ptr PotentialFitPartialChargesAction::performCall()
{
  // check for selected atoms
  const World &world = const_cast<const World &>(World::getInstance());
  if (world.beginAtomSelection() == world.endAtomSelection()) {
    STATUS("There are no atoms selected for fitting partial charges to.");
    return Action::failure;
  }

  /// obtain possible fragments to each selected atom
  AtomFragmentsMap::keysets_t fragments;
  const AtomFragmentsMap::AtomFragmentsMap_t &atommap =
      AtomFragmentsMap::getConstInstance().getMap();
  for (World::AtomSelectionConstIterator iter = world.beginAtomSelection();
      iter != world.endAtomSelection(); ++iter) {
    const AtomFragmentsMap::AtomFragmentsMap_t::const_iterator atomiter =
        atommap.find(iter->first);
    if (atomiter == atommap.end()) {
      ELOG(2, "There are no fragments associated to " << iter->first << ".");
      continue;
    }
    const AtomFragmentsMap::keysets_t &keysets = atomiter->second;
    fragments.insert( fragments.end(), keysets.begin(), keysets.end() );
  }

  /// then go through all fragments and get partial charges for each
  typedef std::map<
      KeySet,
      PartialNucleiChargeFitter::charges_t> FragmentFittedChargeMap_t;
  FragmentFittedChargeMap_t fittedcharges_per_fragment;
  for (AtomFragmentsMap::keysets_t::const_iterator fragmentiter = fragments.begin();
      fragmentiter != fragments.end(); ++fragmentiter) {
    // fragment specifies the homology fragment to use
    SerializablePotential::ParticleTypes_t fragmentnumbers(fragmentiter->begin(), fragmentiter->end());

    // obtain range of homogolous fragments from container
    HomologyContainer &homologies = World::getInstance().getHomologies();
    const HomologyGraph graph = getFirstGraphwithSpecifiedElements(homologies,fragmentnumbers);
    const HomologyContainer::range_t range = homologies.getHomologousGraphs(graph);
    if (range.first == range.second) {
      STATUS("HomologyContainer does not contain specified fragment.");
      return Action::failure;
    }

    // fit and average partial charges over all homologous fragments
    PartialNucleiChargeFitter::charges_t averaged_charges;
    const size_t NoFragments =
        obtainAverageChargesOverFragments(averaged_charges, range, params.radius.get());
    if ((NoFragments == 0) && (range.first != range.second)) {
      STATUS("Fitting for fragment "+toString(*fragmentiter)+" failed.");
      return Action::failure;
    }

    // make sum of charges zero if desired
    if (params.enforceZeroCharge.get())
      enforceZeroTotalCharge(averaged_charges);

    // place into map for later retrieval
    fittedcharges_per_fragment.insert( std::make_pair(*fragmentiter, averaged_charges));

    // output status info fitted charges
    LOG(2, "DEBUG: For fragment " << *fragmentiter << " we have fitted the following charges "
        << averaged_charges << ", averaged over " << NoFragments << " fragments.");
  }

  /// obtain average charge for each atom the fitted charges over all its fragments
  typedef std::map<atomId_t, double> fitted_charges_t;
  fitted_charges_t fitted_charges;
  for (World::AtomSelectionConstIterator atomiter = world.beginAtomSelection();
      atomiter != world.endAtomSelection(); ++atomiter) {
    const AtomFragmentsMap::AtomFragmentsMap_t::const_iterator keysetsiter =
        atommap.find(atomiter->first);
    ASSERT(keysetsiter != atommap.end(),
        "PotentialFitPartialChargesAction::performCall() - we checked already that "
        +toString(atomiter->first)+" should be present!");
    const AtomFragmentsMap::keysets_t & keysets = keysetsiter->second;

    double average_charge = 0.;
    size_t NoFragments = 0;
    // go over all fragments associated to this atom
    for (AtomFragmentsMap::keysets_t::const_iterator keysetsiter = keysets.begin();
        keysetsiter != keysets.end(); ++keysetsiter) {
      const KeySet &keyset = *keysetsiter;

      // find the associated charge in the charge vector
      const FragmentFittedChargeMap_t::const_iterator chargesiter =
          fittedcharges_per_fragment.find(keyset);
      ASSERT(chargesiter != fittedcharges_per_fragment.end(),
          "PotentialFitPartialChargesAction::performCall() - no charge to "+toString(keyset)
          +" any longer present in fittedcharges_per_fragment?");
      const PartialNucleiChargeFitter::charges_t &charges = chargesiter->second;
      PartialNucleiChargeFitter::charges_t::const_iterator chargeiter =
          charges.begin();
      const KeySet::const_iterator keysetiter = keyset.find(atomiter->first);
      ASSERT( keysetiter != keyset.end(),
          "PotentialFitPartialChargesAction::performCall() - atom "
          +toString(atomiter->first)+" not contained in keyset "+toString(keyset));
      std::advance(chargeiter, std::distance(keyset.begin(), keysetiter));
      ASSERT( chargeiter != charges.end(),
          "PotentialFitPartialChargesAction::performCall() - charges "
          +toString(charges.size())+" and keyset "+toString(keyset.size())
          +" do not have the same length?");

      // and add onto charge sum
      const double & charge_in_fragment = *chargeiter;
      average_charge += charge_in_fragment;
      ++NoFragments;
    }
    // average to obtain final partial charge for this atom
    average_charge = 1./(size_t)NoFragments;
    fitted_charges.insert( std::make_pair(atomiter->first, average_charge) );
  }

  /// place all fitted charges into ParticleRegistry
  const ParticleFactory &factory =
      const_cast<const ParticleFactory&>(ParticleFactory::getInstance());
  const periodentafel &periode = *World::getInstance().getPeriode();
  for (fitted_charges_t::const_iterator chargeiter = fitted_charges.begin();
      chargeiter != fitted_charges.end(); ++chargeiter) {
    const atom * const walker = World::getInstance().getAtom(AtomById(chargeiter->first));
    ASSERT( walker != NULL,
        "PotentialFitPartialChargesAction::performCall() - atom "
        +toString(chargeiter->first)+" not present in the World?");
    const double &charge = chargeiter->second;
    const atomicNumber_t elementno = walker->getElementNo();
    const std::string name = Particle::findFreeName(periode, elementno);
    LOG(1, "INFO: Adding particle " << name << " for atom "
        << *walker << " with element " << elementno << ", charge " << charge);
    factory.createInstance(name, elementno, charge);
  }

  return Action::success;
}

ActionState::ptr PotentialFitPartialChargesAction::performUndo(ActionState::ptr _state) {
  return Action::success;
}

ActionState::ptr PotentialFitPartialChargesAction::performRedo(ActionState::ptr _state){
  return Action::success;
}

bool PotentialFitPartialChargesAction::canUndo() {
  return false;
}

bool PotentialFitPartialChargesAction::shouldUndo() {
  return false;
}
/** =========== end of function ====================== */
